/* See the end of this file for copyright and license terms. */

#include <gay/types.h>
#include <gay/util.h>

#include <string.h>

static char *twobyte_strstr(const unsigned char *h, const unsigned char *n)
{
	u16 nw = n[0] << 8 | n[1], hw = h[0] << 8 | h[1];
	for (h++; *h && hw != nw; hw = hw << 8 | *++h)
		/* nothing */;
	return *h ? (char *)h - 1 : 0;
}

static char *threebyte_strstr(const unsigned char *h, const unsigned char *n)
{
	u32 nw = (u32)n[0] << 24 | n[1] << 16 | n[2] << 8;
	u32 hw = (u32)h[0] << 24 | h[1] << 16 | h[2] << 8;
	for (h += 2; *h && hw != nw; hw = (hw | *++h) << 8)
		/* nothing */;
	return *h ? (char *)h - 2 : 0;
}

static char *fourbyte_strstr(const unsigned char *h, const unsigned char *n)
{
	u32 nw = (u32)n[0] << 24 | n[1] << 16 | n[2] << 8 | n[3];
	u32 hw = (u32)h[0] << 24 | h[1] << 16 | h[2] << 8 | h[3];
	for (h += 3; *h && hw != nw; hw = hw << 8 | *++h)
		/* nothing */;
	return *h ? (char *)h - 3 : 0;
}

#define BITOP(a, b, op) \
	((a)[(usize)(b) / (8 * sizeof *(a))] op \
	    (usize)1 << ((usize)(b) % (8 * sizeof *(a))))

/*
 * Two Way string search algorithm, with a bad shift table applied to the last
 * byte of the window. A bit array marks which entries in the shift table are
 * initialized to avoid fully initializing a 1kb/2kb table.
 *
 * Reference: CROCHEMORE M., PERRIN D., 1991, Two-way string-matching,
 * Journal of the ACM 38(3):651-675
 */
static char *twoway_strstr(const unsigned char *h, const unsigned char *n)
{
	const unsigned char *z;
	usize l, ip, jp, k, p, ms, p0, mem, mem0;
	usize byteset[32 / sizeof(usize)] = { 0 };
	usize shift[256];

	/* Computing length of needle and fill shift table */
	for (l = 0; n[l] && h[l]; l++)
		BITOP(byteset, n[l], |=), shift[n[l]] = l + 1;
	if (n[l])
		return 0; /* hit the end of h */

	/* Compute maximal suffix */
	ip = -1;
	jp = 0;
	k = p = 1;
	while (jp + k < l) {
		if (n[ip + k] == n[jp + k]) {
			if (k == p) {
				jp += p;
				k = 1;
			} else
				k++;
		} else if (n[ip + k] > n[jp + k]) {
			jp += k;
			k = 1;
			p = jp - ip;
		} else {
			ip = jp++;
			k = p = 1;
		}
	}
	ms = ip;
	p0 = p;

	/* And with the opposite comparison */
	ip = -1;
	jp = 0;
	k = p = 1;
	while (jp + k < l) {
		if (n[ip + k] == n[jp + k]) {
			if (k == p) {
				jp += p;
				k = 1;
			} else
				k++;
		} else if (n[ip + k] < n[jp + k]) {
			jp += k;
			k = 1;
			p = jp - ip;
		} else {
			ip = jp++;
			k = p = 1;
		}
	}
	if (ip + 1 > ms + 1)
		ms = ip;
	else
		p = p0;

	/* Periodic needle? */
	if (memcmp(n, n + p, ms + 1) != 0) {
		mem0 = 0;
		p = max(ms, l - ms - 1) + 1;
	} else
		mem0 = l - p;
	mem = 0;

	/* Initialize incremental end-of-haystack pointer */
	z = h;

	/* Search loop */
	for (;;) {
		/* Update incremental end-of-haystack pointer */
		if (z - h < l) {
			/* Fast estimate for MIN(l,63) */
			usize grow = l | 63;
			const unsigned char *z2 = memchr(z, 0, grow);
			if (z2) {
				z = z2;
				if (z - h < l)
					return 0;
			} else
				z += grow;
		}

		/* Check last byte first; advance by shift on mismatch */
		if (BITOP(byteset, h[l - 1], &)) {
			k = l - shift[h[l - 1]];
			if (k) {
				if (k < mem)
					k = mem;
				h += k;
				mem = 0;
				continue;
			}
		} else {
			h += l;
			mem = 0;
			continue;
		}

		/* Compare right half */
		for (k = max(ms + 1, mem); n[k] && n[k] == h[k]; k++)
			/* nothing */;
		if (n[k]) {
			h += k - ms;
			mem = 0;
			continue;
		}
		/* Compare left half */
		for (k = ms + 1; k > mem && n[k - 1] == h[k - 1]; k--)
			;
		if (k <= mem)
			return (char *)h;
		h += p;
		mem = mem0;
	}
}

char *strstr(const char *h, const char *n)
{
	/* Return immediately on empty needle */
	if (!n[0])
		return (char *)h;

	/* Use faster algorithms for short needles */
	h = strchr(h, *n);
	if (!h || !n[1])
		return (char *)h;
	if (!h[1])
		return 0;
	if (!n[2])
		return twobyte_strstr((void *)h, (void *)n);
	if (!h[2])
		return 0;
	if (!n[3])
		return threebyte_strstr((void *)h, (void *)n);
	if (!h[3])
		return 0;
	if (!n[4])
		return fourbyte_strstr((void *)h, (void *)n);

	return twoway_strstr((void *)h, (void *)n);
}

/*
 * Copyright (c) 2021 fef <owo@fef.moe>
 * Copyright (c) 2005-2014 Rich Felker, et al.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * $FreeBSD$
 */
